from typing import Optional

from cloud_storage_clients.connector import Connector
from cloud_storage_clients.gcs.client import GcsClient


class GcsClientFactory:
    def __call__(self, connector: Connector, credentials: Optional[dict]):
        return GcsClient(connector, credentials)



import abc

from cloud_storage_clients.connector import Connector


class CredentialRepository(abc.ABC):
    """
    `CredentialRepository` are used by `ClientPool` to retrieve credentials for given `Connector`.
    It should implement get_credentials(connector: Connector) that takes a connector as input
    and return a dict containing credentials used to instantiate a `Client` for example.
    """

    @abc.abstractmethod
    def get_credentials(self, connector: Connector) -> dict:
        pass

    @abc.abstractmethod
    def update_credentials(self, connector: Connector, secret_data: dict) -> None:
        pass

    @abc.abstractmethod
    def delete_credentials(self, connector: Connector) -> None:
        pass



class StorageProviders(Exception):
    pass


class RepositoryError(StorageProviders):
    pass


class ConfigurationError(StorageProviders):
    pass


class ObjectNotFoundError(StorageProviders):
    pass


class FactoryNotFoundError(StorageProviders):
    pass



from cloud_storage_clients.connector import Connector
from cloud_storage_clients.repositories.credentials import CredentialRepository


class ConstantCredentialRepository(CredentialRepository):
    """This class can be used as a CredentialRepository to always return the same credentials"""

    def __init__(self, credentials: dict):
        self.credentials = credentials

    def get_credentials(self, connector: Connector) -> dict:
        return self.credentials

    def update_credentials(self, connector: Connector, secret_data: dict) -> None:
        self.credentials = secret_data

    def delete_credentials(self, connector: Connector) -> None:
        self.credentials = {}



import binascii
import collections
import datetime
import hashlib
import sys
from urllib.parse import quote

import six


class GoogleSigner:
    @staticmethod
    def generate_signed_url(
        google_credentials,
        bucket_name: str,
        object_name: str,
        subresource=None,
        expiration=604800,
        http_method="GET",
        query_parameters=None,
        headers=None,
    ):
        """
        Generate a signed url for a distant resource without getting it first.
        https://cloud.google.com/storage/docs/access-control/signing-urls-manually
        """
        if expiration > 604800:
            print("Expiration Time can't be longer than 604800 seconds (7 days).")
            sys.exit(1)

        escaped_object_name = quote(six.ensure_binary(object_name), safe=b"/~")
        canonical_uri = f"/{escaped_object_name}"

        datetime_now = datetime.datetime.now(tz=datetime.timezone.utc)
        request_timestamp = datetime_now.strftime("%Y%m%dT%H%M%SZ")
        datestamp = datetime_now.strftime("%Y%m%d")

        client_email = google_credentials.service_account_email
        credential_scope = f"{datestamp}/auto/storage/goog4_request"
        credential = f"{client_email}/{credential_scope}"

        if headers is None:
            headers = dict()
        host = f"{bucket_name}.storage.googleapis.com"
        headers["host"] = host

        canonical_headers = ""
        ordered_headers = collections.OrderedDict(sorted(headers.items()))
        for k, v in ordered_headers.items():
            lower_k = str(k).lower()
            strip_v = str(v).lower()
            canonical_headers += f"{lower_k}:{strip_v}\n"

        signed_headers = ""
        for k, _ in ordered_headers.items():
            lower_k = str(k).lower()
            signed_headers += f"{lower_k};"
        signed_headers = signed_headers[:-1]  # remove trailing ';'

        if query_parameters is None:
            query_parameters = dict()
        query_parameters["X-Goog-Algorithm"] = "GOOG4-RSA-SHA256"
        query_parameters["X-Goog-Credential"] = credential
        query_parameters["X-Goog-Date"] = request_timestamp
        query_parameters["X-Goog-Expires"] = expiration
        query_parameters["X-Goog-SignedHeaders"] = signed_headers
        if subresource:
            query_parameters[subresource] = ""

        canonical_query_string = ""
        ordered_query_parameters = collections.OrderedDict(
            sorted(query_parameters.items())
        )
        for k, v in ordered_query_parameters.items():
            encoded_k = quote(str(k), safe="")
            encoded_v = quote(str(v), safe="")
            canonical_query_string += f"{encoded_k}={encoded_v}&"
        canonical_query_string = canonical_query_string[:-1]  # remove trailing '&'

        canonical_request = "\n".join(
            [
                http_method,
                canonical_uri,
                canonical_query_string,
                canonical_headers,
                signed_headers,
                "UNSIGNED-PAYLOAD",
            ]
        )

        canonical_request_hash = hashlib.sha256(canonical_request.encode()).hexdigest()

        string_to_sign = "\n".join(
            [
                "GOOG4-RSA-SHA256",
                request_timestamp,
                credential_scope,
                canonical_request_hash,
            ]
        )

        # signer.sign() signs using RSA-SHA256 with PKCS1v15 padding
        signature = binascii.hexlify(
            google_credentials.signer.sign(string_to_sign)
        ).decode()

        return f"https://{host}{canonical_uri}?{canonical_query_string}&x-goog-signature={signature}"



from typing import Optional, Union

import hvac.adapters
import hvac.exceptions

from cloud_storage_clients.connector import Connector
from cloud_storage_clients.exceptions import RepositoryError
from cloud_storage_clients.repositories import vault_adapter
from cloud_storage_clients.repositories.credentials import CredentialRepository
from cloud_storage_clients.repositories.vault_adapter import DefaultVaultAdapter


class VaultCredentialRepository(CredentialRepository):
    """
    This repository can be used to read credentials through a (Vault)[https://www.vaultproject.io/] secrets manager.
    It needs a VaultAdapter, that you can instantiate and give to constructor, or use method build() to directly
    instantiate this repository.
    A VaultAdapter class should implement at least get_secrets(vault_key: str).

    By default, this repository will build a vault key from given connector id and add a prefix to it.
    """

    def __init__(
        self, adapter: vault_adapter.VaultAdapter, prefix: Optional[str] = None
    ):
        self.prefix = prefix or ""
        self.adapter = adapter

    @staticmethod
    def build(
        url: str,
        *,
        vault_client: Optional[hvac.Client] = None,
        credentials: Union[
            vault_adapter.TokenCredentials, vault_adapter.SecretCredentials, None
        ] = None,
        hvac_client_verify: bool = True,
        secret_mount: Optional[str] = None,
        hvac_adapter: type[hvac.adapters.Adapter] = hvac.adapters.JSONAdapter,
        prefix: Optional[str] = None,
    ):
        adapter = DefaultVaultAdapter(
            url,
            vault_client=vault_client,
            credentials=credentials,
            hvac_client_verify=hvac_client_verify,
            secret_mount=secret_mount,
            hvac_adapter=hvac_adapter,
        )
        return VaultCredentialRepository(adapter, prefix)

    def get_credentials(self, connector: Connector) -> dict:
        vault_key = self.build_vault_key(connector)
        try:
            return self.adapter.get_secrets(vault_key)
        except Exception as e:
            raise RepositoryError(
                f"Could not retrieve credentials of connector {connector.key} from Vault"
            ) from e

    def update_credentials(self, connector: Connector, secret_data: dict) -> dict:
        vault_key = self.build_vault_key(connector)
        try:
            return self.adapter.update_secrets(vault_key, secret_data)
        except Exception as e:
            raise RepositoryError(
                f"Could not update credentials of connector {connector.key} from Vault"
            ) from e

    def delete_credentials(self, connector: Connector) -> dict:
        vault_key = self.build_vault_key(connector)
        try:
            return self.adapter.delete_secrets(vault_key)
        except Exception as e:
            raise RepositoryError(
                f"Could not delete credentials of connector {connector.key} from Vault"
            ) from e

    def build_vault_key(self, connector: Connector):
        vault_path = connector.key

        if self.prefix:
            vault_path = f"{self.prefix}/{vault_path}"

        return vault_path



from datetime import datetime, timedelta
from typing import Any, Optional

from azure.identity import ClientSecretCredential
from azure.storage.blob import (
    BlobClient,
    BlobPrefix,
    BlobSasPermissions,
    BlobServiceClient,
    UserDelegationKey,
    generate_blob_sas,
)
from cloud_storage_clients.client import Client
from cloud_storage_clients.connector import Connector
from cloud_storage_clients.exceptions import ConfigurationError
from cloud_storage_clients.responses import (
    BucketFile,
    BucketFolder,
    BucketObjects,
    CompleteMultipartUploadResponse,
    CopyResponse,
    CreateMultipartUploadResponse,
    DeleteResponse,
    GetDownloadUrlResponse,
    GetPartUploadUrlResponse,
    GetResponse,
    GetUploadUrlResponse,
    HeadResponse,
    ListResponse,
    PutResponse,
)


class AzureClient(Client):
    def __init__(
        self,
        connector: Connector,
        credentials: Optional[dict],
        delegation_key_expiration_time: int,
    ):
        super().__init__(connector, credentials)
        self.delegation_key_expiration_time = delegation_key_expiration_time

        if not credentials:
            raise NotImplementedError()

        if credentials:
            try:
                account_url = credentials["account_url"]
                azure_tenant_id = credentials["azure_tenant_id"]
                azure_client_id = credentials["azure_client_id"]
                azure_client_secret = credentials["azure_client_secret"]
            except KeyError as e:
                raise ConfigurationError() from e

            self._client = BlobServiceClient(
                account_url=account_url,
                credential=ClientSecretCredential(
                    tenant_id=azure_tenant_id,
                    client_id=azure_client_id,
                    client_secret=azure_client_secret,
                ),
            )

    def generate_user_delegation_key(self) -> UserDelegationKey:
        start_time = datetime.utcnow()
        expiry_time = start_time + timedelta(
            seconds=self.delegation_key_expiration_time
        )
        return self._client.get_user_delegation_key(
            key_start_time=start_time, key_expiry_time=expiry_time
        )

    def _get_blob_client(self, object_name: str) -> BlobClient:
        return self._client.get_blob_client(container=self.bucket, blob=object_name)

    def get(self, object_name: str) -> GetResponse:
        blob_client = self._get_blob_client(object_name)
        return GetResponse(content=blob_client.download_blob().read())

    def head(self, object_name: str) -> HeadResponse:
        blob_client = self._get_blob_client(object_name)
        blob = blob_client.get_blob_properties()
        return HeadResponse(
            file_size=blob.size, content_type=blob.content_settings.content_type
        )

    def put(self, object_name: str, data: Any) -> PutResponse:
        blob_client = self._get_blob_client(object_name)
        return PutResponse(
            info=blob_client.upload_blob(data, blob_type="BlockBlob", overwrite=True)
        )

    def delete(self, object_name: str) -> DeleteResponse:
        return DeleteResponse(
            info=self._get_blob_client(object_name).delete_blob(
                delete_snapshots="include"
            )
        )

    def copy(self, source_object_name: str, destination_object_name: str):
        destination_blob = self._client.get_blob_client(
            container=self.bucket, blob=destination_object_name
        )
        source_url = self.get_download_url(
            object_name=source_object_name, expiration=3600
        ).url
        return CopyResponse(info=destination_blob.start_copy_from_url(source_url))

    def list(self, prefix: str, delimiter: str) -> ListResponse:
        folders = []
        files = []
        for blob in self._client.get_container_client(self.bucket).walk_blobs(
            name_starts_with=prefix, delimiter=delimiter
        ):
            if isinstance(blob, BlobPrefix):
                folders.append(BucketFolder(name=blob.name))
            else:
                files.append(
                    BucketFile(
                        name=blob.name, last_modified=blob.creation_time, size=blob.size
                    )
                )

        return ListResponse(bucket_objects=BucketObjects(folders=folders, files=files))

    def _generate_sas_url(self, object_name, expiration, permission) -> str:
        user_delegation_key = self.generate_user_delegation_key()
        blob_client = self._client.get_blob_client(
            container=self.bucket, blob=object_name
        )
        start_time = datetime.utcnow()
        expiry_time = start_time + timedelta(seconds=expiration)
        sas_token = generate_blob_sas(
            account_name=blob_client.account_name,
            container_name=self.bucket,
            blob_name=object_name,
            user_delegation_key=user_delegation_key,
            permission=permission,
            expiry=expiry_time,
            start=start_time,
        )
        return f"{blob_client.url}?{sas_token}"

    def get_download_url(
        self, object_name: str, expiration: int
    ) -> GetDownloadUrlResponse:
        return GetDownloadUrlResponse(
            url=self._generate_sas_url(
                object_name, expiration, BlobSasPermissions(read=True)
            )
        )

    def get_upload_url(
        self, object_name: str, file_type: str, expiration: int
    ) -> GetUploadUrlResponse:
        url = self._generate_sas_url(
            object_name,
            expiration,
            BlobSasPermissions(write=True, read=True, add=True, create=True),
        )
        return GetUploadUrlResponse(url=url, fields={"Content-Type": file_type})

    def create_multipart_upload(
        self, object_name: str
    ) -> CreateMultipartUploadResponse:
        raise NotImplementedError()

    def get_part_upload_url(
        self, object_name: str, upload_id: str, part_number: int, expiration: int
    ) -> GetPartUploadUrlResponse:
        raise NotImplementedError()

    def complete_multipart_upload(
        self, object_name: str, upload_id: str, parts: Any
    ) -> CompleteMultipartUploadResponse:
        raise NotImplementedError()

    def close(self):
        self._client.close()



from datetime import datetime, timedelta
from typing import Any, Optional

from google.cloud import storage
from google.cloud.storage import Blob
from google.oauth2 import service_account

from cloud_storage_clients.client import Client
from cloud_storage_clients.connector import Connector
from cloud_storage_clients.exceptions import ConfigurationError, ObjectNotFoundError
from cloud_storage_clients.gcs.signer import GoogleSigner
from cloud_storage_clients.responses import (
    BucketFile,
    BucketFolder,
    BucketObjects,
    CompleteMultipartUploadResponse,
    CopyResponse,
    CreateMultipartUploadResponse,
    DeleteResponse,
    GetDownloadUrlResponse,
    GetPartUploadUrlResponse,
    GetResponse,
    GetUploadUrlResponse,
    HeadResponse,
    ListResponse,
    PutResponse,
)


class GcsClient(Client):
    def __init__(self, connector: Connector, credentials: Optional[dict]):
        super().__init__(connector, credentials)

        if not credentials:
            raise NotImplementedError()

        try:
            private_key = credentials["private_key"]
            self._client_email = credentials["client_email"]
        except KeyError as e:
            raise ConfigurationError() from e

        account_info = {
            "type": "service_account",
            "private_key": private_key.replace("\\n", "\n"),
            "client_email": self._client_email,
            "token_uri": "https://oauth2.googleapis.com/token",
        }
        self._google_credentials = (
            service_account.Credentials.from_service_account_info(account_info)
        )
        self._client = storage.Client(project="", credentials=self._google_credentials)
        self.google_bucket = self._client.get_bucket(self.bucket)

    def _get_blob(self, object_name) -> Blob:
        return self.google_bucket.get_blob(object_name)

    def get(self, object_name: str) -> GetResponse:
        blob = self._get_blob(object_name)
        if not blob:
            raise ObjectNotFoundError(f"Object {object_name} not found")
        return GetResponse(content=blob.download_as_bytes())

    def head(self, object_name: str) -> HeadResponse:
        blob = self._get_blob(object_name)
        if not blob:
            raise ObjectNotFoundError(f"Object {object_name} not found")
        return HeadResponse(file_size=blob.size, content_type=blob.content_type)

    def put(self, object_name: str, data: Any) -> PutResponse:
        blob = self.google_bucket.blob(object_name)
        blob.upload_from_file(data)
        return PutResponse(info={"name": blob.name, "size": blob.size})

    def delete(self, object_name: str) -> DeleteResponse:
        blob = self._get_blob(object_name)
        if not blob:
            raise ObjectNotFoundError(f"Object {object_name} not found")
        return DeleteResponse(info=blob.delete())

    def copy(
        self, source_object_name: str, destination_object_name: str
    ) -> CopyResponse:
        source_blob = self._get_blob(source_object_name)
        if not source_blob:
            raise ObjectNotFoundError(f"Object {source_object_name} not found")
        return self.google_bucket.copy_blob(
            source_blob, self.google_bucket, new_name=destination_object_name
        )

    def list(self, prefix: str, delimiter: str) -> ListResponse:
        blobs = self.google_bucket.list_blobs(prefix=prefix, delimiter=delimiter)
        files = [
            BucketFile(name=blob.name, last_modified=blob.updated, size=blob.size)
            for blob in blobs
            if not blob.name.endswith("/")
        ]
        folders = [BucketFolder(name=prefix) for prefix in blobs.prefixes]
        return ListResponse(bucket_objects=BucketObjects(folders=folders, files=files))

    def get_download_url(
        self, object_name: str, expiration: int
    ) -> GetDownloadUrlResponse:
        signed_url = GoogleSigner.generate_signed_url(
            self._google_credentials, self.bucket, object_name, expiration=expiration
        )
        return GetDownloadUrlResponse(url=signed_url)

    def get_upload_url(
        self, object_name: str, file_type: str, expiration: int
    ) -> GetUploadUrlResponse:
        expiration_date = datetime.utcnow() + timedelta(seconds=expiration)
        presigned_post = self._client.generate_signed_post_policy_v4(
            bucket_name=self.bucket,
            blob_name=object_name,
            fields={"Content-Type": file_type},
            conditions=[{"Content-Type": file_type}],
            expiration=expiration_date,
        )
        return GetUploadUrlResponse(
            url=presigned_post["url"],
            fields=presigned_post["fields"],
        )

    def create_multipart_upload(
        self, object_name: str
    ) -> CreateMultipartUploadResponse:
        raise NotImplementedError()

    def get_part_upload_url(
        self, object_name: str, upload_id: str, part_number: int, expiration: int
    ) -> GetPartUploadUrlResponse:
        raise NotImplementedError()

    def complete_multipart_upload(
        self, object_name: str, upload_id: str, parts: Any
    ) -> CompleteMultipartUploadResponse:
        raise NotImplementedError()

    def close(self):
        self._client.close()



import abc
import logging
from dataclasses import dataclass
from typing import Optional, Union

import hvac
import hvac.adapters
import hvac.exceptions


class VaultAdapter(abc.ABC):
    """This class can be implemented to connect to a Vault secrets manager.
    It will be used by a VaultCredentialRepository to access and retrieve secrets holding connectors credentials.
    """

    @abc.abstractmethod
    def get_secrets(self, vault_key: str):
        pass

    @abc.abstractmethod
    def update_secrets(self, vault_key: str, secret_data: dict):
        pass

    @abc.abstractmethod
    def delete_secrets(self, vault_key: str):
        pass


@dataclass
class SecretCredentials:
    role_id: str
    secret_id: str


@dataclass
class TokenCredentials:
    token: str
    unseal_key: str


class DefaultVaultAdapter(VaultAdapter):
    def __init__(
        self,
        url: str,
        *,
        vault_client: Optional[hvac.Client] = None,
        credentials: Union[SecretCredentials, TokenCredentials, None] = None,
        hvac_client_verify: bool = True,
        secret_mount: Optional[str] = None,
        hvac_adapter: type[hvac.adapters.Adapter] = hvac.adapters.JSONAdapter,
    ):
        self.url = url
        self.hvac_client_verify = hvac_client_verify
        self.secret_mount = secret_mount or ""

        if vault_client:
            self.client = vault_client
        elif isinstance(credentials, SecretCredentials):
            self.client = self.get_hvac_client_from_secret(
                url, credentials, hvac_adapter, hvac_client_verify
            )
        elif isinstance(credentials, TokenCredentials):
            self.client = self.get_hvac_client_from_token(
                url, credentials, hvac_adapter, hvac_client_verify
            )
        else:
            raise RuntimeError(
                "Please give a vault client or credentials as VaultSecret or VaultToken"
            )

        if not self.client.is_authenticated():
            raise RuntimeError("Vault Client not authenticated.")
        else:
            logging.debug("Vault client initialized")

    @staticmethod
    def get_hvac_client_from_secret(
        url: str,
        credentials: SecretCredentials,
        hvac_adapter: type[hvac.adapters.Adapter],
        hvac_client_verify: bool,
    ):
        client = hvac.Client(
            url=url,
            verify=hvac_client_verify,
            adapter=hvac_adapter,
            kwargs={
                "role_id": credentials.role_id,
                "secret_id": credentials.secret_id,
            },
        )
        client.auth.approle.login(
            role_id=credentials.role_id,
            secret_id=credentials.secret_id,
        )
        return client

    @staticmethod
    def get_hvac_client_from_token(
        url: str,
        credentials: TokenCredentials,
        hvac_adapter: type[hvac.adapters.Adapter],
        hvac_client_verify: bool,
    ):
        client = hvac.Client(
            url=url,
            token=credentials.token,
            verify=hvac_client_verify,
            adapter=hvac_adapter,
        )

        if client.sys.is_sealed():
            client.sys.submit_unseal_key(credentials.unseal_key)

        return client

    def get_secrets(self, vault_key: str) -> dict:
        return self.client.secrets.kv.v2.read_secret(
            path=vault_key, mount_point=self.secret_mount
        )["data"]["data"]

    def update_secrets(self, vault_key: str, secret_data: dict):
        self.client.secrets.kv.v2.create_or_update_secret(
            path=vault_key,
            secret=secret_data,
            mount_point=self.secret_mount,
        )

    def delete_secrets(self, vault_key: str):
        self.client.secrets.kv.v2.delete_metadata_and_all_versions(
            path=vault_key,
            mount_point=self.secret_mount,
        )



from typing import Callable, Optional

from cloud_storage_clients.azure.factory import AzureClientFactory
from cloud_storage_clients.cache import ClientCache, DictCache
from cloud_storage_clients.client import Client
from cloud_storage_clients.connector import Connector
from cloud_storage_clients.exceptions import FactoryNotFoundError
from cloud_storage_clients.gcs.factory import GcsClientFactory
from cloud_storage_clients.repositories.credentials import CredentialRepository
from cloud_storage_clients.s3.factory import S3ClientFactory

ClientFactory = Callable[[Connector, Optional[dict]], Client]

DEFAULT_CLIENT_FACTORIES = {
    "azure": AzureClientFactory(3600),
    "google": GcsClientFactory(),
    "s3": S3ClientFactory(),
}


class ClientPool:
    def __init__(
        self,
        default_repository: CredentialRepository = None,
        client_factories: dict[str, ClientFactory] = None,
        cache: ClientCache = None,
    ):
        self.default_repository = default_repository
        self.client_factories = client_factories or DEFAULT_CLIENT_FACTORIES
        self.cache = cache or DictCache()

    def get_client(
        self,
        connector: Connector,
        repository: Optional[CredentialRepository] = None,
        override_cache: bool = False,
    ) -> Client:
        if self.cache and not override_cache:
            client = self.cache.get(connector)
            if client:
                return client

        try:
            factory = self.client_factories[connector.client_type]
        except KeyError:
            raise FactoryNotFoundError(
                f"No factory found for client type {connector.client_type}"
            )

        credentials = self.get_credentials(connector, repository)

        client = factory(connector, credentials)

        if self.cache:
            self.cache.set(connector, client)

        return client

    def register_factory(self, client_type: str, factory: ClientFactory):
        self.client_factories[client_type] = factory

    def get_credentials(
        self, connector: Connector, repository: Optional[CredentialRepository]
    ):
        if repository:
            return repository.get_credentials(connector)
        elif self.default_repository:
            return self.default_repository.get_credentials(connector)
        else:
            return None

    def reset_cache(self, connector: Connector):
        if self.cache:
            self.cache.delete(connector)



from typing import Optional

from cloud_storage_clients.azure.client import AzureClient
from cloud_storage_clients.connector import Connector


class AzureClientFactory:
    def __init__(self, delegation_key_expiration_time: int):
        self.delegation_key_expiration_time = delegation_key_expiration_time

    def __call__(self, connector: Connector, credentials: Optional[dict]):
        return AzureClient(connector, credentials, self.delegation_key_expiration_time)



from dataclasses import dataclass
from typing import Optional


@dataclass
class Connector:
    client_type: str
    bucket: str
    id: Optional[str] = None

    @property
    def key(self):
        if self.id:
            return self.id
        else:
            return f"{self.client_type}/{self.bucket}"



from dataclasses import dataclass
from datetime import datetime
from typing import Optional


@dataclass
class GetResponse:
    content: bytes


@dataclass
class HeadResponse:
    file_size: int
    content_type: Optional[str]


@dataclass
class PutResponse:
    info: Optional[dict]


@dataclass
class DeleteResponse:
    info: Optional[dict]


@dataclass
class CopyResponse:
    info: Optional[dict]


@dataclass
class BucketFolder:
    name: str


@dataclass
class BucketFile:
    name: str
    last_modified: datetime
    size: int


@dataclass
class BucketObjects:
    folders: list[BucketFolder]
    files: list[BucketFile]


@dataclass
class ListResponse:
    bucket_objects: BucketObjects


@dataclass
class GetDownloadUrlResponse:
    url: str


@dataclass
class GetUploadUrlResponse:
    url: str
    fields: dict


@dataclass
class CreateMultipartUploadResponse:
    upload_id: str


@dataclass
class GetPartUploadUrlResponse:
    url: str


@dataclass
class CompleteMultipartUploadResponse:
    info: Optional[dict]



import os
from typing import Any, Optional

import boto3
from botocore.config import Config

from cloud_storage_clients.client import Client
from cloud_storage_clients.connector import Connector
from cloud_storage_clients.exceptions import ConfigurationError, ObjectNotFoundError
from cloud_storage_clients.responses import (
    BucketFile,
    BucketFolder,
    BucketObjects,
    CompleteMultipartUploadResponse,
    CopyResponse,
    CreateMultipartUploadResponse,
    DeleteResponse,
    GetDownloadUrlResponse,
    GetPartUploadUrlResponse,
    GetResponse,
    GetUploadUrlResponse,
    HeadResponse,
    ListResponse,
    PutResponse,
)


class S3Client(Client):
    def __init__(
        self,
        connector: Connector,
        credentials: Optional[dict],
        default_region_name: Optional[str],
    ):
        super().__init__(connector, credentials)

        if not credentials:
            raise NotImplementedError()

        try:
            access_key_id = credentials["access_key_id"]
            secret_access_key = credentials["secret_access_key"]
        except KeyError as e:
            raise ConfigurationError() from e

        if "region_name" in credentials:
            region_name = credentials["region_name"]
        else:
            region_name = default_region_name

        if "endpoint_url" in credentials:
            endpoint_url = credentials["endpoint_url"]
        else:
            endpoint_url = f"https://s3.{region_name}.amazonaws.com"

        self._client = boto3.client(
            "s3",
            endpoint_url=endpoint_url,
            region_name=region_name,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            config=Config(signature_version="s3v4"),
        )

    def get(self, object_name: str) -> GetResponse:
        return GetResponse(
            content=self._client.get_object(Bucket=self.bucket, Key=object_name)[
                "Body"
            ].read()
        )

    def head(self, object_name: str) -> HeadResponse:
        metadata = self._client.head_object(Bucket=self.bucket, Key=object_name)
        if not metadata:
            raise ObjectNotFoundError()
        return HeadResponse(
            file_size=metadata["ContentLength"], content_type=metadata["ContentType"]
        )

    def put(self, object_name: str, data: Any) -> PutResponse:
        return PutResponse(
            info=self._client.put_object(Bucket=self.bucket, Key=object_name, Body=data)
        )

    def delete(self, object_name: str) -> DeleteResponse:
        return DeleteResponse(
            info=self._client.delete_object(Bucket=self.bucket, Key=object_name)
        )

    def copy(
        self, source_object_name: str, destination_object_name: str
    ) -> CopyResponse:
        copy_source = os.path.join(self.bucket, source_object_name)
        return CopyResponse(
            info=self._client.copy_object(
                Bucket=self.bucket,
                CopySource=copy_source,
                Key=destination_object_name,
            )
        )

    def list(self, prefix: str, delimiter: str) -> ListResponse:
        paginator = self._client.get_paginator("list_objects_v2")
        folders = []
        files = []
        for page in paginator.paginate(
            Bucket=self.bucket, Prefix=prefix, Delimiter=delimiter
        ):
            if "CommonPrefixes" in page:
                folders.extend(
                    [
                        BucketFolder(name=folder["name"])
                        for folder in page["CommonPrefixes"]
                    ]
                )
            if "Contents" in page:
                files.extend(
                    [
                        BucketFile(
                            name=file["Key"],
                            last_modified=file["LastModified"],
                            size=file["Size"],
                        )
                        for file in page["Contents"]
                    ]
                )
        return ListResponse(bucket_objects=BucketObjects(folders=folders, files=files))

    def get_download_url(
        self, object_name: str, expiration: int
    ) -> GetDownloadUrlResponse:
        return GetDownloadUrlResponse(
            url=self._client.generate_presigned_url(
                "get_object",
                Params={
                    "Bucket": self.bucket,
                    "Key": object_name,
                },
                ExpiresIn=expiration,
            )
        )

    def get_upload_url(
        self, object_name: str, file_type: str, expiration: int
    ) -> GetUploadUrlResponse:
        presigned_post = self._client.generate_presigned_post(
            Bucket=self.bucket,
            Key=object_name,
            Fields={"Content-Type": file_type},
            Conditions=[{"Content-Type": file_type}],
            ExpiresIn=expiration,
        )
        return GetUploadUrlResponse(
            url=presigned_post["url"],
            fields=presigned_post["fields"],
        )

    def create_multipart_upload(
        self, object_name: str
    ) -> CreateMultipartUploadResponse:
        response = self._client.create_multipart_upload(
            Bucket=self.bucket, Key=object_name
        )
        return CreateMultipartUploadResponse(upload_id=response["UploadId"])

    def get_part_upload_url(
        self, object_name: str, upload_id: str, part_number: int, expiration: int
    ) -> GetPartUploadUrlResponse:
        return GetPartUploadUrlResponse(
            url=self._client.generate_presigned_url(
                ClientMethod="upload_part",
                Params={
                    "Bucket": self.bucket,
                    "Key": object_name,
                    "UploadId": upload_id,
                    "PartNumber": part_number,
                },
                ExpiresIn=expiration,
            )
        )

    def complete_multipart_upload(
        self, object_name: str, upload_id: str, parts: Any
    ) -> CompleteMultipartUploadResponse:
        return CompleteMultipartUploadResponse(
            info=self._client.complete_multipart_upload(
                Bucket=self.bucket,
                Key=object_name,
                MultipartUpload={"Parts": parts},
                UploadId=upload_id,
            )
        )

    def close(self):
        self._client.close()



from typing import Optional

from cloud_storage_clients.connector import Connector
from cloud_storage_clients.s3.client import S3Client


class S3ClientFactory:
    def __init__(self, default_region_name: Optional[str] = None):
        self.default_region_name = default_region_name

    def __call__(self, connector: Connector, credentials: Optional[dict]):
        return S3Client(connector, credentials, self.default_region_name)



import abc
from abc import abstractmethod
from typing import Any, Optional

from cloud_storage_clients.connector import Connector
from cloud_storage_clients.responses import (
    CompleteMultipartUploadResponse,
    CopyResponse,
    CreateMultipartUploadResponse,
    DeleteResponse,
    GetDownloadUrlResponse,
    GetPartUploadUrlResponse,
    GetResponse,
    GetUploadUrlResponse,
    HeadResponse,
    ListResponse,
    PutResponse,
)


class Client(abc.ABC):
    def __init__(self, connector: Connector, credentials: Optional[dict]):
        self.connector = connector
        self.bucket = connector.bucket
        self.is_authenticated = credentials is not None

    @abstractmethod
    def get(self, object_name: str) -> GetResponse:
        pass

    @abstractmethod
    def head(self, object_name: str) -> HeadResponse:
        pass

    @abstractmethod
    def put(self, object_name: str, data: Any) -> PutResponse:
        pass

    @abstractmethod
    def delete(self, object_name: str) -> DeleteResponse:
        pass

    @abstractmethod
    def copy(
        self, source_object_name: str, destination_object_name: str
    ) -> CopyResponse:
        pass

    @abstractmethod
    def list(self, prefix: str, delimiter: str) -> ListResponse:
        pass

    @abstractmethod
    def get_download_url(
        self, object_name: str, expiration: int
    ) -> GetDownloadUrlResponse:
        pass

    @abstractmethod
    def get_upload_url(
        self, object_name: str, file_type: str, expiration: int
    ) -> GetUploadUrlResponse:
        pass

    @abstractmethod
    def create_multipart_upload(
        self, object_name: str
    ) -> CreateMultipartUploadResponse:
        pass

    @abstractmethod
    def get_part_upload_url(
        self, object_name: str, upload_id: str, part_number: int, expiration: int
    ) -> GetPartUploadUrlResponse:
        pass

    @abstractmethod
    def complete_multipart_upload(
        self, object_name: str, upload_id: str, parts: Any
    ) -> CompleteMultipartUploadResponse:
        pass

    @abstractmethod
    def close(self):
        pass



import abc
import logging
from threading import Lock
from typing import Optional

from cloud_storage_clients.client import Client
from cloud_storage_clients.connector import Connector


class ClientCache:
    @abc.abstractmethod
    def get(self, key: Connector) -> Optional[Client]:
        pass

    @abc.abstractmethod
    def set(self, key: Connector, value: Client, timeout: int = None):
        pass

    @abc.abstractmethod
    def delete(self, key: Connector):
        pass

    @abc.abstractmethod
    def clear(self):
        pass


class DictCache(ClientCache):
    clients: dict[str, Client]

    def __init__(self) -> None:
        self.clients = {}
        self.lock = Lock()

    def get_cache_key(self, connector: Connector):
        return connector.key

    def get(self, connector: Connector) -> Optional[Client]:
        with self.lock:
            cache_key = self.get_cache_key(connector)
            if cache_key in self.clients:
                return self.clients[cache_key]

            return None

    def set(self, connector: Connector, client: Client, timeout: int = None):
        with self.lock:
            cache_key = self.get_cache_key(connector)
            self.clients[cache_key] = client

    def delete(self, connector: Connector):
        with self.lock:
            cache_key = self.get_cache_key(connector)
            if cache_key in self.clients:
                try:
                    self.clients[cache_key].close()
                except Exception:
                    logging.exception(
                        f"Could not close client from connector {cache_key}"
                    )
                del self.clients[cache_key]

    def clear(self):
        with self.lock:
            for key, client in self.clients.items():
                try:
                    client.close()
                except Exception:
                    logging.exception(f"Could not close client from connector {key}")

            self.clients.clear()


class DisabledCache(ClientCache):
    def get(self, connector: Connector) -> Optional[Client]:
        return None

    def set(self, connector: Connector, client: Client, timeout: int = None):
        pass

    def delete(self, connector: Connector):
        pass

    def clear(self):
        pass



import os

from cloud_storage_clients.connector import Connector
from cloud_storage_clients.repositories.credentials import CredentialRepository


class EnvCredentialRepository(CredentialRepository):
    """
    EnvCredentialRepository is a repository that reads environment variables and parse them into a dict of credentials.
    It uses the client_type of a connector to look for default expected environment variable of known cloud storages.
    """

    AZURE_CLIENT_TYPES = ["azure"]
    GCS_CLIENT_TYPES = ["gcs", "google"]
    S3_CLIENT_TYPES = ["s3", "aws"]

    AZURE_ACCOUNT_URL = "ACCOUNT_URL"
    AZURE_CLIENT_ID = "CLIENT_ID"
    AZURE_TENANT_ID = "TENANT_ID"
    AZURE_CLIENT_SECRET = "CLIENT_SECRET"

    GCS_PRIVATE_KEY = "PRIVATE_KEY"
    GCS_CLIENT_EMAIL = "CLIENT_EMAIL"

    S3_ACCESS_KEY_ID = "ACCESS_KEY_ID"
    S3_SECRET_ACCESS_KEY = "SECRET_ACCESS_KEY"
    S3_ENDPOINT_URL = "ENDPOINT_URL"
    S3_REGION_NAME = "REGION_NAME"

    def __init__(self, prefix: str = ""):
        self.prefix = prefix

    def get_key(self, env):
        return self.prefix + env

    def getenv(self, env):
        key = self.get_key(env)
        var = os.getenv(key)
        if var is None:
            raise KeyError(f"Variable {key} not found in environment.")
        return var

    def setenv(self, env: str, value: str):
        key = self.get_key(env)
        os.environ[key] = value

    def delenv(self, env):
        key = self.get_key(env)
        try:
            del os.environ[key]
        except KeyError:
            pass

    def get_credentials(self, connector: Connector) -> dict:
        if connector.client_type.lower() in self.S3_CLIENT_TYPES:
            return {
                "access_key_id": self.getenv(self.S3_ACCESS_KEY_ID),
                "secret_access_key": self.getenv(self.S3_SECRET_ACCESS_KEY),
                "region_name": self.getenv(self.S3_REGION_NAME),
                "endpoint_url": self.getenv(self.S3_ENDPOINT_URL),
            }
        elif connector.client_type.lower() in self.AZURE_CLIENT_TYPES:
            return {
                "account_url": self.getenv(self.AZURE_ACCOUNT_URL),
                "client_id": self.getenv(self.AZURE_CLIENT_ID),
                "tenant_id": self.getenv(self.AZURE_TENANT_ID),
                "client_secret": self.getenv(self.AZURE_CLIENT_SECRET),
            }
        elif connector.client_type.lower() in self.GCS_CLIENT_TYPES:
            return {
                "private_key": self.getenv(self.GCS_PRIVATE_KEY),
                "client_email": self.getenv(self.GCS_CLIENT_EMAIL),
            }
        raise NotImplementedError("unknown connector type")

    @staticmethod
    def _assert_keys_coherence(secrets: dict, expected_secret_keys: set):
        secret_keys = set(secrets.keys())
        if not expected_secret_keys.issubset(secret_keys):
            raise KeyError(
                f"Some keys are missing: {expected_secret_keys - secret_keys}"
            )

    def update_credentials(self, connector: Connector, secret_data: dict) -> None:
        if connector.client_type.lower() in self.AZURE_CLIENT_TYPES:
            self._assert_keys_coherence(
                secret_data, {"account_url", "client_id", "tenant_id", "client_secret"}
            )
            self.setenv(self.AZURE_ACCOUNT_URL, secret_data["account_url"])
            self.setenv(self.AZURE_CLIENT_ID, secret_data["client_id"])
            self.setenv(self.AZURE_TENANT_ID, secret_data["tenant_id"])
            self.setenv(self.AZURE_CLIENT_SECRET, secret_data["client_secret"])
        elif connector.client_type.lower() in self.GCS_CLIENT_TYPES:
            self._assert_keys_coherence(secret_data, {"private_key", "client_email"})
            self.setenv(self.GCS_PRIVATE_KEY, secret_data["private_key"])
            self.setenv(self.GCS_CLIENT_EMAIL, secret_data["client_email"])
        elif connector.client_type.lower() in self.S3_CLIENT_TYPES:
            self._assert_keys_coherence(
                secret_data,
                {"access_key_id", "secret_access_key", "region_name", "endpoint_url"},
            )

            self.setenv(self.S3_ACCESS_KEY_ID, secret_data["access_key_id"])
            self.setenv(self.S3_SECRET_ACCESS_KEY, secret_data["secret_access_key"])
            self.setenv(self.S3_REGION_NAME, secret_data["region_name"])
            self.setenv(self.S3_ENDPOINT_URL, secret_data["endpoint_url"])
        else:
            raise NotImplementedError("unknown connector type")

    def delete_credentials(self, connector: Connector) -> None:
        if connector.client_type.lower() in self.S3_CLIENT_TYPES:
            self.delenv(self.S3_ACCESS_KEY_ID)
            self.delenv(self.S3_SECRET_ACCESS_KEY)
            self.delenv(self.S3_REGION_NAME)
            self.delenv(self.S3_ENDPOINT_URL)
        elif connector.client_type.lower() in self.AZURE_CLIENT_TYPES:
            self.delenv(self.AZURE_ACCOUNT_URL)
            self.delenv(self.AZURE_CLIENT_ID)
            self.delenv(self.AZURE_TENANT_ID)
            self.delenv(self.AZURE_CLIENT_SECRET)
        elif connector.client_type.lower() in self.GCS_CLIENT_TYPES:
            self.delenv(self.GCS_PRIVATE_KEY)
            self.delenv(self.GCS_CLIENT_EMAIL)
        else:
            raise NotImplementedError("unknown connector type")
